//*****************************************************************************
//
// Datei       : bitvecto.cc
// Autor       : Markus Halla
//
// Funktion    : Definiert eine Klasse fr eine Mengenreprsentation mit einem
//               Bit-Vektor.
//
//*****************************************************************************

#include <math.h>
#include <string.h>
#include "const.h"
#include "bitvecto.h"

//*****************************************************************************
// Klasse      : BitVector
// Methode     : Konstruktor
//
// Funktion    : Initialisieren der Menge
// Parameter   : newSize - die Anzahl der Elemente die verarbeitet werden
//                         knnen
// Ergebnis    : ---
//
//*****************************************************************************
BitVector::BitVector(const SetElement newSize)
    : size(newSize), posInit(IteratorStart)
{
    dataSize     = long(ceil(newSize / (double)bitsInElement));
    data         = new unsigned char[dataSize] = 0;
};

//*****************************************************************************
// Klasse      : BitVector
// Methode     : Copy-Konstruktor
//
// Funktion    : Initialisieren der Menge mit einer anderen
// Parameter   : rValue - Referenz auf die zu kopierende Menge
// Ergebnis    : ---
//
//*****************************************************************************
BitVector::BitVector(const BitVector& rValue)
    : size(rValue.size), pos(rValue.pos), posInit(rValue.posInit)
{
    dataSize     = rValue.dataSize;
    data         = new unsigned char[dataSize];
    memcpy(data, rValue.data, sizeof(unsigned char) * dataSize);
};

//*****************************************************************************
// Klasse      : BitVector
// Methode     : Destruktor
//
// Funktion    : Lschen der Menge
// Parameter   : ---
// Ergebnis    : ---
//
//*****************************************************************************
BitVector::~BitVector()
{
    delete[] data;
    dataSize = 0;
};

//*****************************************************************************
// Klasse      : BitVector
// Methode     : Zuweisungsoperator
//
// Funktion    : Weist der Menge die Daten von "rValue" zu
// Parameter   : rValue - Referenz auf die zu kopierende Menge
// Ergebnis    : ---
//
//*****************************************************************************
const BitVector& BitVector::operator=(const BitVector& rValue)
{
    // Prfen auf Selbstzuweisung
    if (this != &rValue) {
        // Wenn notwendig Speicher fr neue Daten erzeugen
        if (dataSize != rValue.dataSize) {
            delete[] data;
            data     = new unsigned char[dataSize];
            dataSize = rValue.dataSize;
        }

        // Die Daten kopieren
        size     = rValue.size;
        pos      = rValue.pos;
        posInit  = rValue.posInit;
        memcpy(data, rValue.data, sizeof(unsigned char) * dataSize);
    }

    return *this;
};

//*****************************************************************************
// Klasse      : BitVector
// Methode     : clearSet
//
// Funktion    : Lscht alle Elemente aus der Menge
// Parameter   : ---
// Ergebnis    : ---
//
//*****************************************************************************
void BitVector::clearSet()
{
    // Alle Elemente lschen
    memset(data, 0, sizeof(unsigned char) * dataSize);
};

//*****************************************************************************
// Klasse      : BitVector
// Methode     : chooseOne
//
// Funktion    : Whlt ein Element aus der Menge aus und liefert dessen Nummer
//               zurck (das erste Element wird gewhlt)
// Parameter   : ---
// Ergebnis    : Nummer des gewhlten Elemente
//               -1 wenn kein Element in der Menge ist
//
//*****************************************************************************
SetElement BitVector::chooseOne() const
{
    SetElement i = 0;   // Laufvariable

    // Alle Elemente durchsuchen
    while ((i < dataSize) && (data[i] == 0))
        i++;

    // Haben wir alle Datenelemente durchsucht?
    if (i == dataSize)
        return -1;

    // Das erste Element in diesem Datenelement suchen
    for (short j = 0; j < bitsInElement; j++) {
        if ((data[i] & (1 << j)) != 0)
            return j + i * bitsInElement;
    }

    // Hierher drften wir nicht kommen
    return(-1);
}

//*****************************************************************************
// Klasse      : BitVector
// Methode     : cardinality
//
// Funktion    : Ermittelt die Anzahl der Elemente in der Menge
// Parameter   : ---
// Ergebnis    : Anzahl der Elemente in der Menge
//
//*****************************************************************************
SetElement BitVector::cardinality() const
{
    SetElement sum = 0;  // Zwischensumme

    // Alle Elemente durchsuchen
    for (SetElement i = 0; i < size; i++) {
        if (member(i))
            sum++;
    }

    return sum;
}

//*****************************************************************************
// Klasse      : BitVector
// Methode     : operator== - Vergleichsoperator
//
// Funktion    : Vergleicht die beiden Menge
// Parameter   : rValue - Referenz auf die zu vergleichende Menge
// Ergebnis    : Boolscher Wert - Vergleichsergebnis
//
//*****************************************************************************
bool BitVector::operator==(const BitVector& rValue) const
{
    // Sind die beiden Mengen gleich gro?
    if (size != rValue.size)
        return false;

    // Alle Datenelemente einzeln vergleichen
    if (memcmp(data, rValue.data, sizeof(unsigned char) * dataSize) == 0)
        return true;
    else
        return false;
}

//*****************************************************************************
// Klasse      : BitVector
// Methode     : operator| - Vereinigungsoperator
//
// Funktion    : Liefert die Vereinigung der beiden Menge
// Parameter   : rValue - Referenz auf die zu vergleichende Menge
// Ergebnis    : Referenz auf den Bit-Vektor
// Exception   : E_Size - Unterschiedliche Gre der beiden Mengen
//
//*****************************************************************************
BitVector& BitVector::operator|(const BitVector& rValue) const
{
    // Sind die beiden Mengen gleich gro?
    if (size != rValue.size)
        ERRORSIZE

    // Den Bit-Vektor fr das Ergebnis erzeugen
    BitVector* result = new BitVector(size);

    // Alle Datenelemente einzeln verknpfen
    for (SetElement i = 0; i < dataSize; i++)
        result->data[i] = data[i] | rValue.data[i];

    return *result;
}

//*****************************************************************************
// Klasse      : BitVector
// Methode     : vereinigung
//
// Funktion    : Wandelt die eigene Menge in die Vereinigung der beiden Mengen
//               um
// Parameter   : rValue1 - Die erste Parametermenge
//               rValue2 - Die zweite Parametermenge
// Ergebnis    : ---
// Exception   : E_Size - Unterschiedliche Gre der beiden Mengen
//
//*****************************************************************************
void BitVector::vereinigung(const BitVector& rValue1, const BitVector& rValue2)
{
    // Sind die beiden Mengen gleich gro?
    if (size != rValue1.size != rValue2.size)
        ERRORSIZE

    // Alle Datenelemente einzeln verknpfen
    for (SetElement i = 0; i < dataSize; i++)
        data[i] = rValue1.data[i] | rValue2.data[i];
}

//*****************************************************************************
// Klasse      : BitVector
// Methode     : operator& - Durchschnittsoperator
//
// Funktion    : Liefert die Durchschnitt der beiden Menge
// Parameter   : rValue - Referenz auf die zu vergleichende Menge
// Ergebnis    : Referenz auf den Bit-Vektor
// Exception   : E_Size - Unterschiedliche Gre der beiden Mengen
//
//*****************************************************************************
BitVector& BitVector::operator&(const BitVector& rValue) const
{
    // Sind die beiden Mengen gleich gro?
    if (size != rValue.size)
        ERRORSIZE

    // Den Bit-Vektor fr das Ergebnis erzeugen
    BitVector* result = new BitVector(size);

    // Alle Datenelemente einzeln verknpfen
    for (SetElement i = 0; i < dataSize; i++)
        result->data[i] = data[i] & rValue.data[i];

    return *result;
}

//*****************************************************************************
// Klasse      : BitVector
// Methode     : intersect
//
// Funktion    : Wandelt die eigene Menge in den Durchschnitt der beiden Mengen
//               um
// Parameter   : rValue1 - Die erste Parametermenge
//               rValue2 - Die zweite Parametermenge
// Ergebnis    : ---
// Exception   : E_Size - Unterschiedliche Gre der beiden Mengen
//
//*****************************************************************************
void BitVector::intersect(const BitVector& rValue1, const BitVector& rValue2)
{
    // Sind die beiden Mengen gleich gro?
    if (size != rValue1.size != rValue2.size)
        ERRORSIZE

    // Alle Datenelemente einzeln verknpfen
    for (SetElement i = 0; i < dataSize; i++)
        data[i] = rValue1.data[i] & rValue2.data[i];
}

//*****************************************************************************
// Klasse      : BitVector
// Methode     : operator^ - Differenzoperator
//
// Funktion    : Liefert die Differenz der beiden Menge
// Parameter   : rValue - Referenz auf die zu vergleichende Menge
// Ergebnis    : Referenz auf den Bit-Vektor
// Exception   : E_Size - Unterschiedliche Gre der beiden Mengen
//
//*****************************************************************************
BitVector& BitVector::operator^(const BitVector& rValue) const
{
    // Sind die beiden Mengen gleich gro?
    if (size != rValue.size)
        ERRORSIZE

    // Den Bit-Vektor fr das Ergebnis erzeugen
    BitVector* result = new BitVector(size);

    // Alle Datenelemente einzeln verknpfen
    for (SetElement i = 0; i < dataSize; i++)
        result->data[i] = data[i] ^ rValue.data[i];

    return *result;
}

//*****************************************************************************
// Klasse      : BitVector
// Methode     : difference
//
// Funktion    : Wandelt die eigene Menge in die Differenz der beiden Mengen um
// Parameter   : rValue1 - Die erste Parametermenge
//               rValue2 - Die zweite Parametermenge
// Ergebnis    : ---
// Exception   : E_Size - Unterschiedliche Gre der beiden Mengen
//
//*****************************************************************************
void BitVector::difference(const BitVector& rValue1, const BitVector& rValue2)
{
    // Sind die beiden Mengen gleich gro?
    if (size != rValue1.size != rValue2.size)
        ERRORSIZE

    // Alle Datenelemente einzeln verknpfen
    for (SetElement i = 0; i < dataSize; i++)
        data[i] = rValue1.data[i] ^ rValue2.data[i];
}

//*****************************************************************************
// Klasse      : BitVector
// Methode     : operator! - Komplementoperator
//
// Funktion    : Liefert das Komplement der eigenen Menge
// Parameter   : ---
// Ergebnis    : Referenz auf den Bit-Vektor
//
//*****************************************************************************
BitVector& BitVector::operator!() const
{
    // Den Bit-Vektor fr das Ergebnis erzeugen
    BitVector* result = new BitVector(size);

    // Alle Datenelemente einzeln bearbeiten
    for (SetElement i = 0; i < dataSize; i++)
        result->data[i] = ~data[i];

    return *result;
}

//*****************************************************************************
// Klasse      : BitVector
// Methode     : complement
//
// Funktion    : Wandelt die eigene Menge in das Komplement der "rValue" Menge
//               um
// Parameter   : rValue - Die Parametermenge
// Ergebnis    : ---
// Exception   : E_Size - Unterschiedliche Gre der beiden Mengen
//
//*****************************************************************************
void BitVector::complement(const BitVector& rValue)
{
    // Sind die beiden Mengen gleich gro?
    if (size != rValue.size)
        ERRORSIZE

    // Alle Datenelemente einzeln bearbeiten
    for (SetElement i = 0; i < dataSize; i++)
        data[i] = ~rValue.data[i];
}

//*****************************************************************************
// Klasse      : BitVector
// Methode     : setStart
//
// Funktion    : Setzt den Iterator an den Anfang der Menge.
// Parameter   : ---
// Ergebnis    : ---
//
//*****************************************************************************
void BitVector::setStart()
{
    posInit = IteratorStart;
}

//*****************************************************************************
// Klasse      : BitVector
// Methode     : getNext
//
// Funktion    : Die nchste Elementnummer ermitteln und zurckgeben.
// Parameter   : ---
// Ergebnis    : Nummer der nchsten Elements
//               -1 - Kein weiteres Element vorhanden
//
//*****************************************************************************
SetElement BitVector::getNext()
{
    // Sind wir am Ende der Menge angelangt
    if (posInit == IteratorEnd)
        return -1;

    // Es ist das erste Element der Menge das wir suchen?
    if (posInit == IteratorStart) {
        pos     = 0;
        posInit = IteratorWork;
    // Ansonsten beginnen wir beim nchsten Element zu suchen
    } else
        pos++;

    // Der Iterator ist in Arbeit

    // Laufvariable
    SetElement i = pos / bitsInElement;

    // Gibt es noch ein Element in diesem Datenelement
    for (short j = pos % bitsInElement; j < bitsInElement; j++) {
        if ((data[i] & (1 << j)) != 0) {
            pos = j + i * bitsInElement;
            return pos;
        }
    }

    // Alle weiteren Elemente durchsuchen
    do {
        i++;
    } while ((i < dataSize) && (data[i] == 0));

    // Haben wir alle Datenelemente durchsucht?
    if (i == dataSize) {
        posInit = IteratorEnd;
        return -1;
    }

    // Das erste Element in diesem Datenelement suchen
    for (short j = 0; j < bitsInElement; j++) {
        if ((data[i] & (1 << j)) != 0) {
            pos = j + i * bitsInElement;
            return pos;
        }
    }

    // Hierher drften wir nicht kommen
    posInit = IteratorEnd;
    return -1;
}

//*****************************************************************************
// Klasse      : BitVector
// Methode     : print
//
// Funktion    : Gibt die Elemente der Menge aus
// Parameter   : ---
// Ergebnis    : ---
//
//*****************************************************************************
#include <stdio.h>
void BitVector::print(FILE *fp)
{
    // Alle Datenelemente einzeln bearbeiten
    for (SetElement i = 0; i < size; i++) {
        if (member(i))
            fprintf(fp, "%d ", i);
    }

    fprintf(fp, "\n");
}

