#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <values.h>
#include "const.h"
#include "bitvecto.h"

//---------------------------------------------------------------------------
//                                                                 Konstanten
//---------------------------------------------------------------------------
#define ZEILENMAX 100
#define ERROR(text) { printf("%s\n", text); exit(1); }

//---------------------------------------------------------------------------
//                                                            Typdefinitionen
//---------------------------------------------------------------------------
// Ein 2D-Vektor
typedef struct {
    float x, y;    // Die Koordinaten des Startpunkt
    float dx, dy;  // Die relativen Koordinaten des Endpunkt
} T_Vektor_2D;

// Eine Wand des Labyrinths
typedef struct {
    T_Vektor_2D vektor;       // Die transformierte Wand in Vektordarstellung
} T_WandLab;

//---------------------------------------------------------------------------
//                                                          Globale Variablen
//---------------------------------------------------------------------------
T_WandLab   *waende;
int         waendeAnzahl;
char        G_Zeile[ZEILENMAX];
int         maxX, maxY;
int         sizeX, sizeY;
BitVector  *voxel;

//---------------------------------------------------------------------------
//                                                             schneideStrahl
//---------------------------------------------------------------------------
// Parameter:    strahl - Vektor der den Blickstrahl beschreibt
//               wand   - Vektor der die Wand beschreibt
// Rckgabewert: Flag, ob sich der Strahl und die Wand schneiden.
//
// Beschreibung: Berechnet den Schnittpunkt der beiden Vektoren und prft,
//               ob der Strahl die Wand getroffen hat.
//
//---------------------------------------------------------------------------
int schneideStrahl(T_Vektor_2D *strahl, T_Vektor_2D *wand, float *ls)
{
    float d;       // Die Determinanten des Gleichungssystems
    float a1, a2;  // Parameter im Gleichungssystem
    float lw;

    // Berechne den Schnittpunkt. Lsen des Gleichungssystems
    d  = (strahl->dy * wand->dx) - (strahl->dx * wand->dy);

    // Gibt es berhaupt einen Schnittpunkt?
    if (d != 0) {
        a1  = strahl->x - wand->x;
        a2  = strahl->y - wand->y;
        *ls = ((a1 * wand->dy) - (a2 * wand->dx)) / d;
        lw  = ((strahl->dy * a1) - (strahl->dx * a2)) / d;
    } else
        return false;

    // Prfen, ob der Schnittpunkt auerhalb der Wand oder hinter der Kamera liegt.
    if ((lw < 0.0) || (lw > 1.0) || (*ls < 0.0))
        return false;

    return true;
}

//---------------------------------------------------------------------------
//                                                            checkIdentifier
//---------------------------------------------------------------------------
// Parameter:    fp   - Filepointer der Labyrinthdatei
//               text - Identifier der in der Datei vorkommen mu
// Rckgabewert: ---
//
// Beschreibung: Liest bis zum Auftreten des Identifiers Zeilen ein.
//               Wenn er nicht gefunden wird -> Fehlermeldung.
//
//---------------------------------------------------------------------------
void checkIdentifier(FILE *fp, char *text)
{
    while (!feof(fp)) {
        if (fgets(G_Zeile, ZEILENMAX, fp) == NULL)
            ERROR("Lesefehler bei der Labyrinthdatei!")
        if (strcmp(G_Zeile, text) == 0)
            return;
    }

    ERROR("Konnte in der Labyrinthdatei den gewnschten Identifier nicht finden!")
}

//---------------------------------------------------------------------------
//                                                              initLabWaende
//---------------------------------------------------------------------------
// Parameter:    fp - Filepointer der Labyrinthdatei
// Rckgabewert: ---
//
// Beschreibung: Bearbeitet die Wnde des Labyrinths in der Labyrinthdatei.
//
//---------------------------------------------------------------------------
void initLabWaende(FILE *fp)
{
    int   i;              // Schleifenvariable
    float x1, x2, y1, y2; // Die Eckpunkte der Wand
    int   tInt;
    float tFloat;

    // Das Vorhandensein der Labyrint-Waendedaten prfen
    checkIdentifier(fp, "##LABYRINTH-WAENDE\n");

    // Die Anzahl der Wnde einlesen
    if (fscanf(fp, "%d", &waendeAnzahl) != 1)
        ERROR("Falsches Zeilenformat in Labyrinthdatei - Labyrinthwnde!")

    // Speicher fr die Daten besorgen
    if ((waende = (T_WandLab *)calloc(waendeAnzahl, sizeof(T_WandLab))) == NULL)
        ERROR("Nicht gengend Speicher fr die Labyrinthwnde!")

    // Die Wandhhe einlesen
    if (fscanf(fp, "%f", &tFloat) != 1)
        ERROR("Falsches Zeilenformat in Labyrinthdatei - Labyrinthwnde!")

    // Die Datenzeilen einlesen
    for (i = 0; i < waendeAnzahl; i++) {
        // Einen Wand einlesen
        if (fscanf(fp, "%f %f %f %f %f %d", &x1, &y1, &x2, &y2, &tFloat, &tInt) != 6)
            ERROR("Falsches Zeilenformat in Labyrinthdatei - Labyrinthwnde!")

        // Und in die Struktur eintragen
        waende[i].vektor.x  = x1;
        waende[i].vektor.y  = y1;
        waende[i].vektor.dx = x2 - x1;
        waende[i].vektor.dy = y2 - y1;
    }
}

//---------------------------------------------------------------------------
//                                                                 checkPoint
//---------------------------------------------------------------------------
void checkPoint(float x, float y, int vX, int vY)
{
    T_Vektor_2D strahl;
    int         wnr;
    float       tiefe, tiefeNeu;
    int         wand;
    float       winkel;

    strahl.x = x;
    strahl.y = y;

    for (winkel = 0.0; winkel < 360.0; winkel++) {
        strahl.dx = sin(winkel / 180.0 * PI);
        strahl.dy = cos(winkel / 180.0 * PI);
        tiefe = MAXFLOAT;
        for (wand = 0; wand < waendeAnzahl; wand++) {
            if (schneideStrahl(&strahl, &waende[wand].vektor, &tiefeNeu)) {
                // Ist die neue Wand nher?
                if ((tiefeNeu < tiefe) && (tiefeNeu > 0.0)) {
                    wnr   = wand;
                    tiefe = tiefeNeu;
                }
            }
        }

        // Die Wand in die Menge aufnehmen
        voxel[vY * sizeY + vX].addMember(wnr);
    }
}

//---------------------------------------------------------------------------
//                                                                       main
//---------------------------------------------------------------------------
int main(int argc, char *argv[])
{
    FILE       *fpIn;          // Filepointer der Labyrinthdatei
    FILE       *fpOut;         // Filepointer der Labyrinthdatei
    char        fileIn[100];
    char        fileOut[100];
    int         x, y;
    int         anzahl;

    // Programmidentifikation
    printf("Sichtbarkeitsberechnung fr Roswitha-Labyrinthe\n");
    printf("(C) 1999, Markus Halla\n\n");

    // Den Namen der Labyrinthdatei einlesen
    printf("Geben sie den Namen der Labyrinthdatei ein:\n");
    scanf("%s", fileIn);

    // Den Namen der Ergebnissdatei einlesen
    printf("Geben sie den Namen der Voxeldatei ein:\n");
    scanf("%s", fileOut);

    // Die Datei zum Lesen ffnen
    if ((fpIn = fopen(fileIn, "r")) == NULL)
        ERROR("Kann die Labyrinthdatei nicht ffnen!")

    // Den Identifikationsstring berprfen
    if (fgets(G_Zeile, ZEILENMAX, fpIn) == NULL)
        ERROR("Lesefehler bei der Labyrinthdatei!")
    if (strcmp(G_Zeile, "ROSWITHA-LABYRINTHDEFINITION V1.0\n") != 0)
        ERROR("Die Datei ist keine gltige Labyrinthdatei!")

    // Die einzelnen Abschnitte bearbeiten
    printf("Lese die Labyrinthbeschreibung ein\n");
    initLabWaende(fpIn);

    // Die Datei schlieen
    fclose(fpIn);

    // Die Gre des Labyrinths einlesen
    printf("Breite des Labyrinths (0 - x):\n");
    scanf("%d", &maxX);
    printf("Hhe des Labyrinths (0 - y):\n");
    scanf("%d", &maxY);
    sizeX = maxX + 1;
    sizeY = maxY + 1;

    // Die Bitvektoren initialisieren
    voxel = new BitVector[sizeX * sizeY](waendeAnzahl);

    // Die sichtbaren Wnde von jedem Voxel aus ermitteln
    for (x = 0; x < sizeX; x++) {
        for (y = 0; y < sizeY; y++) {
            printf("Berechne Voxel %d, %d\n", x, y);

            checkPoint(x, y, x, y);
            checkPoint(x + 0.5, y, x, y);
            checkPoint(x + 1.0, y, x, y);
            checkPoint(x      , y + 0.5, x, y);
            checkPoint(x + 0.5, y + 0.5, x, y);
            checkPoint(x + 1.0, y + 0.5, x, y);
            checkPoint(x      , y + 1.0, x, y);
            checkPoint(x + 0.5, y + 1.0, x, y);
            checkPoint(x + 1.0, y + 1.0, x, y);
        }
    }

    // Die Datei zum Schreiben ffnen
    if ((fpOut = fopen(fileOut, "w")) == NULL)
        ERROR("Kann die Voxeldatei nicht ffnen!")

    // Den Identifikationsstring schreiben
    fprintf(fpOut, "##LABYRINTH-VOXEL\n");

    // Die Dimensionen der Labyrinthvoxel schreiben
    fprintf(fpOut, "%d %d\n", sizeX, sizeY);

    // Die gesamte Gre der Voxeleintrge berechnen
    anzahl = 0;
    for (x = 0; x < sizeX; x++) {
        for (y = 0; y < sizeY; y++) {
            anzahl += (voxel[y * sizeY + x].cardinality());
        }
    }

    // Die Gesamtanzahl der Voxeleintrge ausgeben
    fprintf(fpOut, "%d\n", anzahl);

    // Die einzelnen Voxel ausgeben
    for (x = 0; x < sizeX; x++) {
        for (y = 0; y < sizeY; y++) {
            fprintf(fpOut, "%d  ", voxel[y * sizeY + x].cardinality());
            voxel[y * sizeY + x].print(fpOut);
        }
    }

    // Die Datei schlieen
    fclose(fpIn);

    // Die Bitvektoren freigeben
    delete[] voxel;
};

